/*******************************************************************************
 * Copyright (c) 2021 Microsoft Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * Contributors:
 *     Microsoft Corporation - initial API and implementation
*******************************************************************************/
package org.eclipse.jdt.ls.core.internal.commands;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.ls.core.internal.WorkspaceHelper;
import org.eclipse.jdt.ls.core.internal.managers.AbstractInvisibleProjectBasedTest;
import org.eclipse.lsp4j.Position;
import org.eclipse.lsp4j.SymbolKind;
import org.eclipse.lsp4j.TextDocumentIdentifier;
import org.eclipse.lsp4j.legacy.typeHierarchy.ResolveTypeHierarchyItemParams;
import org.eclipse.lsp4j.legacy.typeHierarchy.TypeHierarchyDirection;
import org.eclipse.lsp4j.legacy.typeHierarchy.TypeHierarchyItem;
import org.eclipse.lsp4j.legacy.typeHierarchy.TypeHierarchyParams;
import org.junit.Before;
import org.junit.Test;

public class TypeHierarchyCommandTest extends AbstractInvisibleProjectBasedTest {

	private IProject fJProject;
	private TypeHierarchyCommand fCommand;

	@Before
	public void setup() throws Exception {
		importProjects("maven/salut");
		fJProject = WorkspaceHelper.getProject("salut");
		fCommand = new TypeHierarchyCommand();
	}

	@Test
	public void testTypeHierarchy() throws Exception {
		IProgressMonitor monitor = new NullProgressMonitor();
		TypeHierarchyParams params = new TypeHierarchyParams();
		String uriString = fJProject.getFile("src/main/java/org/sample/TestJavadoc.java").getLocationURI().toString();
		TextDocumentIdentifier identifier = new TextDocumentIdentifier(uriString);
		Position position = new Position(4, 20);
		params.setTextDocument(identifier);
		params.setResolve(1);
		params.setDirection(TypeHierarchyDirection.Both);
		params.setPosition(position);
		TypeHierarchyItem item = fCommand.typeHierarchy(params, monitor);
		assertNotNull(item);
		assertEquals(item.getName(), "TestJavadoc");
		assertNotNull(item.getChildren());
		assertEquals(item.getChildren().size(), 0);
		assertNotNull(item.getParents());
		assertEquals(item.getParents().size(), 1);
		assertEquals(item.getParents().get(0).getName(), "Object");
	}

	@Test
	public void testSuperTypeHierarchy() throws Exception {
		IProgressMonitor monitor = new NullProgressMonitor();
		TypeHierarchyParams params = new TypeHierarchyParams();
		String uriString = fJProject.getFile("src/main/java/org/sample/CallHierarchy.java").getLocationURI().toString();
		TextDocumentIdentifier identifier = new TextDocumentIdentifier(uriString);
		Position position = new Position(7, 27);
		params.setTextDocument(identifier);
		params.setResolve(1);
		params.setDirection(TypeHierarchyDirection.Parents);
		params.setPosition(position);
		TypeHierarchyItem item = fCommand.typeHierarchy(params, monitor);
		assertNotNull(item);
		assertEquals(item.getName(), "CallHierarchy$FooBuilder");
		assertNull(item.getChildren());
		assertEquals(item.getParents().size(), 2);
		TypeHierarchyItem builder = item.getParents().get(0);
		assertNotNull(builder);
		assertEquals(builder.getName(), "Builder");
		assertNull(builder.getParents());
		TypeHierarchyItem object = item.getParents().get(1);
		assertNotNull(object);
		assertEquals(object.getName(), "Object");
		assertNull(object.getParents());
	}

	@Test
	public void testSubTypeHierarchy() throws Exception {
		IProgressMonitor monitor = new NullProgressMonitor();
		TypeHierarchyParams params = new TypeHierarchyParams();
		String uriString = fJProject.getFile("src/main/java/org/sample/CallHierarchy.java").getLocationURI().toString();
		TextDocumentIdentifier identifier = new TextDocumentIdentifier(uriString);
		Position position = new Position(2, 43);
		params.setTextDocument(identifier);
		params.setResolve(2);
		params.setDirection(TypeHierarchyDirection.Children);
		params.setPosition(position);
		TypeHierarchyItem item = fCommand.typeHierarchy(params, monitor);
		assertNotNull(item);
		assertEquals(item.getName(), "Builder");
		assertNull(item.getParents());
		assertEquals(item.getChildren().size(), 9);
		for (TypeHierarchyItem child : item.getChildren()) {
			List<TypeHierarchyItem> subChild = child.getChildren();
			assertNotNull(subChild);
			if (subChild.size() == 1) {
				assertEquals(subChild.get(0).getName(), "ReflectionToStringBuilder");
			}
		}
	}

	// https://github.com/redhat-developer/vscode-java/issues/2871
	@Test
	public void testMultipleProjects() throws Exception {
		importProjects("eclipse/gh2871");
		IProject project = WorkspaceHelper.getProject("project1");
		IProgressMonitor monitor = new NullProgressMonitor();
		TypeHierarchyParams params = new TypeHierarchyParams();
		String uriString = project.getFile("src/org/sample/First.java").getLocationURI().toString();
		TextDocumentIdentifier identifier = new TextDocumentIdentifier(uriString);
		Position position = new Position(1, 22);
		params.setTextDocument(identifier);
		params.setResolve(1);
		params.setDirection(TypeHierarchyDirection.Both);
		params.setPosition(position);
		TypeHierarchyItem item = fCommand.typeHierarchy(params, monitor);
		assertNotNull(item);
		assertEquals(item.getName(), "First");
		assertNotNull(item.getChildren());
		assertEquals(item.getChildren().size(), 1);
		assertEquals(item.getChildren().get(0).getName(), "Second");
	}

	@Test
	public void testMethodHierarchy() throws Exception {
		importProjects("maven/type-hierarchy");
		IProject project = WorkspaceHelper.getProject("type-hierarchy");
		String uriString = project.getFile("src/main/java/org/example/Zero.java").getLocationURI().toString();
		TextDocumentIdentifier identifier = new TextDocumentIdentifier(uriString);
		Position position = new Position(3, 17); // public void f[o]o()
		TypeHierarchyParams zeroParams = new TypeHierarchyParams();
		zeroParams.setTextDocument(identifier);
		zeroParams.setResolve(1);
		zeroParams.setDirection(TypeHierarchyDirection.Both);
		zeroParams.setPosition(position);
		TypeHierarchyItem zero = fCommand.typeHierarchy(zeroParams, monitor);

		// do not show java.lang.Object if target method isn't from there
		assertEquals(0, zero.getParents().size());

		assertEquals(SymbolKind.Class, zero.getKind()); // zero
		assertEquals(SymbolKind.Class, zero.getChildren().get(0).getKind()); // one
		assertEquals(SymbolKind.Null, zero.getChildren().get(1).getKind()); // two

		ResolveTypeHierarchyItemParams oneParams = new ResolveTypeHierarchyItemParams();
		oneParams.setItem(zero.getChildren().get(0));
		oneParams.setDirection(TypeHierarchyDirection.Both);
		oneParams.setResolve(1);
		TypeHierarchyItem one = fCommand.resolveTypeHierarchy(oneParams, monitor);

		assertEquals(SymbolKind.Null, one.getChildren().get(1).getKind()); // three
		assertEquals(SymbolKind.Class, one.getChildren().get(0).getKind()); // four

		ResolveTypeHierarchyItemParams twoParams = new ResolveTypeHierarchyItemParams();
		twoParams.setItem(zero.getChildren().get(1));
		twoParams.setDirection(TypeHierarchyDirection.Both);
		twoParams.setResolve(1);
		TypeHierarchyItem two = fCommand.resolveTypeHierarchy(twoParams, monitor);

		assertEquals(SymbolKind.Null, two.getChildren().get(0).getKind()); // five
		assertEquals(SymbolKind.Class, two.getChildren().get(1).getKind()); // six
	}
}
